@@ -19,7 +19,13 @@
 
 namespace Doctrine\DBAL\Portability;
 
-use PDO;
+use Doctrine\DBAL\Driver\StatementIterator;
+use Doctrine\DBAL\FetchMode;
+use Doctrine\DBAL\ParameterType;
+use function array_change_key_case;
+use function is_null;
+use function is_string;
+use function rtrim;
 
 /**
  * Portability wrapper for a Statement.
@@ -31,7 +37,7 @@ use PDO;
 class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
 {
     /**
-     * @var integer
+     * @var int
      */
     private $portability;
 
@@ -41,14 +47,14 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
     private $stmt;
 
     /**
-     * @var integer
+     * @var int
      */
     private $case;
 
     /**
-     * @var integer
+     * @var int
      */
-    private $defaultFetchMode = PDO::FETCH_BOTH;
+    private $defaultFetchMode = FetchMode::MIXED;
 
     /**
      * Wraps <tt>Statement</tt> and applies portability measures.
@@ -66,15 +72,15 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
     /**
      * {@inheritdoc}
      */
-    public function bindParam($column, &$variable, $type = null, $length = null)
+    public function bindParam($column, &$variable, $type = ParameterType::STRING, $length = null)
     {
         return $this->stmt->bindParam($column, $variable, $type, $length);
     }
+
     /**
      * {@inheritdoc}
      */
-
-    public function bindValue($param, $value, $type = null)
+    public function bindValue($param, $value, $type = ParameterType::STRING)
     {
         return $this->stmt->bindValue($param, $value, $type);
     }
@@ -134,24 +140,24 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
      */
     public function getIterator()
     {
-        $data = $this->fetchAll();
-
-        return new \ArrayIterator($data);
+        return new StatementIterator($this);
     }
 
     /**
      * {@inheritdoc}
      */
-    public function fetch($fetchMode = null)
+    public function fetch($fetchMode = null, $cursorOrientation = \PDO::FETCH_ORI_NEXT, $cursorOffset = 0)
     {
         $fetchMode = $fetchMode ?: $this->defaultFetchMode;
 
         $row = $this->stmt->fetch($fetchMode);
 
-        $row = $this->fixRow($row,
-            $this->portability & (Connection::PORTABILITY_EMPTY_TO_NULL|Connection::PORTABILITY_RTRIM),
-            !is_null($this->case) && ($fetchMode == PDO::FETCH_ASSOC || $fetchMode == PDO::FETCH_BOTH) && ($this->portability & Connection::PORTABILITY_FIX_CASE)
-        );
+        $iterateRow = $this->portability & (Connection::PORTABILITY_EMPTY_TO_NULL|Connection::PORTABILITY_RTRIM);
+        $fixCase    = $this->case !== null
+            && ($fetchMode === FetchMode::ASSOCIATIVE || $fetchMode === FetchMode::MIXED)
+            && ($this->portability & Connection::PORTABILITY_FIX_CASE);
+
+        $row = $this->fixRow($row, $iterateRow, $fixCase);
 
         return $row;
     }
@@ -159,25 +165,28 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
     /**
      * {@inheritdoc}
      */
-    public function fetchAll($fetchMode = null, $columnIndex = 0)
+    public function fetchAll($fetchMode = null, $fetchArgument = null, $ctorArgs = null)
     {
         $fetchMode = $fetchMode ?: $this->defaultFetchMode;
 
-        if ($columnIndex != 0) {
-            $rows = $this->stmt->fetchAll($fetchMode, $columnIndex);
+        if ($fetchArgument) {
+            $rows = $this->stmt->fetchAll($fetchMode, $fetchArgument);
         } else {
             $rows = $this->stmt->fetchAll($fetchMode);
         }
 
         $iterateRow = $this->portability & (Connection::PORTABILITY_EMPTY_TO_NULL|Connection::PORTABILITY_RTRIM);
-        $fixCase = !is_null($this->case) && ($fetchMode == PDO::FETCH_ASSOC || $fetchMode == PDO::FETCH_BOTH) && ($this->portability & Connection::PORTABILITY_FIX_CASE);
+        $fixCase    = $this->case !== null
+            && ($fetchMode === FetchMode::ASSOCIATIVE || $fetchMode === FetchMode::MIXED)
+            && ($this->portability & Connection::PORTABILITY_FIX_CASE);
+
         if ( ! $iterateRow && !$fixCase) {
             return $rows;
         }
 
-        if ($fetchMode === PDO::FETCH_COLUMN) {
+        if ($fetchMode === FetchMode::COLUMN) {
             foreach ($rows as $num => $row) {
-                $rows[$num] = array($row);
+                $rows[$num] = [$row];
             }
         }
 
@@ -185,7 +194,7 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
             $rows[$num] = $this->fixRow($row, $iterateRow, $fixCase);
         }
 
-        if ($fetchMode === PDO::FETCH_COLUMN) {
+        if ($fetchMode === FetchMode::COLUMN) {
             foreach ($rows as $num => $row) {
                 $rows[$num] = $row[0];
             }
@@ -195,9 +204,9 @@ class Statement implements \IteratorAggregate, \Doctrine\DBAL\Driver\Statement
     }
 
     /**
-     * @param mixed   $row
-     * @param integer $iterateRow
-     * @param boolean $fixCase
+     * @param mixed $row
+     * @param int   $iterateRow
+     * @param bool  $fixCase
      *
      * @return array
      */
